
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "config.h"
#include "libavutil/common.h"
#include "mp_trans_msg.h"


#define MODE_NONE	0
#define MODE_DEFINE	1

#define ARGS_END	"diouxXeEfFgGaAcspn%"

typedef void *vptr;

typedef struct {
    char *msgid;
    unsigned long int hash;
    unsigned int length;
    unsigned int offset;
    char *msgstr;
    void *next;
    } msgitem;

typedef struct {
    char *defname;
    char *defvalue;
    int  line;
    void *next;
    } defitem;

msgitem *rootitem=NULL;
msgitem *lastitem=NULL;
defitem *master_def=NULL;
defitem *trans_def=NULL;
int itemno=0;
int ifs[1024];
int ifnumber=0;

char msgid[8192];
char msgstr[8192];

/* We assume to have `unsigned long int' value with at least 32 bits.  */
#define HASHWORDBITS 32


/* Defines the so called `hashpjw' function by P.J. Weinberger
   [see Aho/Sethi/Ullman, COMPILERS: Principles, Techniques and Tools,
   1986, 1987 Bell Telephone Laboratories, Inc.]  */

static unsigned long int hash_string (const char *text)
{
  unsigned long int hval, g;
  const char *str = text;

  /* Compute the hash value for the given string.  */
  hval = 0;
  while (*str != '\0')
    {
      hval <<= 4;
      hval += (unsigned long int) *str++;
      g = hval & ((unsigned long int) 0xf << (HASHWORDBITS - 4));
      if (g != 0)
	{
	  hval ^= g >> (HASHWORDBITS - 8);
	  hval ^= g;
	}
    }
  return hval;
}

void push_if(int value) {
if(ifnumber==1023) return;
ifs[++ifnumber]=value;
}

void pop_if() {
if (ifnumber==0) return;
ifnumber--;
}

void invert_if() {
if (ifs[ifnumber]) ifs[ifnumber]=1; else ifs[ifnumber]=0;
}

int isif() {
    return ifs[ifnumber];
    }

char* trim(char *txt)
  {
  int i,l;
  char *ret;
  if (strlen(txt)==0) return txt;
  ret=txt;
  l=strlen(ret);
  for(i=0;i<l;i++)
    {
    if ((txt[i]==32) || (txt[i]==9))
      {
      ret++;
      }
      else i=l;
    }
  l=strlen(ret);
  for(i=1;i<l;i++)
    {
    if ((ret[strlen(ret)-1]==32) || (ret[strlen(ret)-1]==9))
      {
      ret[strlen(ret)-1]=0;
      }
    }
  if (txt!=ret) strcpy(txt,ret);
  return txt;  
  }

char* getdefname(char *bp,char **defname) {
    char *pt,*ps;
    bp=trim(bp);
    ps=strchr(bp,' ');
    pt=strchr(bp,'\t');
    if(pt==NULL) pt=ps;
    if(ps==NULL) ps=pt;
    if(ps && pt && pt>ps) ps=pt;
    if(ps) *ps++=0;
    if(*defname) free(*defname);
    *defname=strdup(bp);
    return ps;
}

char* seek_def(defitem *def, char *defname) {
    if(!def) return NULL;
    if(!defname) return NULL;
    while (def) {
	if (strcmp(defname,def->defname)==0 && strlen(defname)==strlen(def->defname)) return def->defvalue;
	def=def->next;
	}
    return NULL;
}

defitem* store_def(defitem *def,char *defname,char *defvalue, int line) {
    defitem *ret;
    if (!defname) return def;
    if (NULL==(ret=malloc(sizeof(defitem)))) return def;
    ret->defname=strdup(defname);
    ret->line=line;
    if (defvalue) ret->defvalue=strdup(defvalue); else ret->defvalue=strdup("");
    ret->next=def;
    return ret;
}

char* conv_c_format(char *text) {
int i,n=0,a=0,farg=0;
char nc;
const int l = strlen(text);
char *args;
if (l==0) return NULL;
if (NULL==(args=malloc(l))) return NULL;
memset(args,0,l);
for(i=0;i<l;i++) {
    if (!farg && text[i]=='%') farg=1; else
	if (farg && strchr(ARGS_END,text[i])) farg=2;
    if (farg) {args[a]=text[i]; a++;}
    if (farg==2) farg=0;
    nc=text[i];
    if (text[i]=='\\')
	switch (text[i+1]) {
	    case '"' : nc='"'; i++; break;
	    case 'a' : nc='\a'; i++; break;
	    case 'b' : nc='\b'; i++; break;
	    case 'f' : nc='\t'; i++; break;
	    case 'n' : nc='\n'; i++; break;
	    case 'r' : nc='\r'; i++; break;
	    case 't' : nc='\t'; i++; break;
	    case 'v' : nc='\v'; i++; break;
	    case '\'' : nc='\''; i++; break;
	    case '\\' : nc='\\'; i++; break;
	    }
    text[n]=nc;
    n++;
    }
text[n]='\0';
return args;
}


int proc_str(char *text, char *dst, int line, defitem *def) {
char *start, *end, *p, *newstr;
if (!text || strlen(text)==0) return 0;
do {
    end=NULL;
    if (NULL!=(start=strchr(text,'"'))) {
	*start=0;
	start++;
	p=start;
	end=NULL;
	text=trim(text);
	if (strlen(text)>0 && NULL!=(newstr=seek_def(def,text))) {
	    strncat(dst,newstr,8192);
	    }
	do {
	    if (NULL==(p=strchr(p,'"'))) end=p; else
		{
		if (*(p-1)!='\\') {
		    *p=0;
		    end=++p;
		    }
		p++;
		}
	    } while (end==NULL);
	}
    if (start) strncat(dst,start,8192);
    text=end;
    } while (text!=NULL);
return 0;
}

void proc_msgid(int line, char *defname) {
msgitem *item;
char *args_msgid=conv_c_format(msgid);
char *args_msgstr=conv_c_format(msgstr);
if (!args_msgid && !args_msgstr) return;
if (!args_msgid || !args_msgstr) return;
if (strlen(args_msgid)+strlen(args_msgstr)==0 || (strcmp(args_msgid,args_msgstr)==0 && strlen(args_msgid)==strlen(args_msgstr))) {
    if (NULL==(item=malloc(sizeof(msgitem)))) {printf("Memory allocation error.\n"); return;}
    item->msgid=strdup(msgid);
    item->hash=hash_string(msgid);
    item->length=strlen(msgid);
    item->msgstr=strdup(msgstr);
    item->next=NULL;
    if (lastitem) {
	lastitem->next=item;
	lastitem=item;
	} else {
	rootitem=item; 
	lastitem=item;
	}
    itemno++;
    }
    else
    {
    printf("Error: different arguments: %i line. (%s) '%s-%s'\n",line,defname,args_msgstr,msgstr);
    }
free(args_msgid); free(args_msgstr);
return;
}

int proc_include(char *fname,defitem **def) {
FILE *fh;
int mode=MODE_NONE, line=0, msgline=0, endmode=0, defmode=0;
char buffer[1024];
char *bp, *ht, lineend;
char *defname=NULL, *ifdefname=NULL;
memset(ifs,0,sizeof(ifs));
ifnumber=0;
ifs[0]=1;

if(NULL==(fh=fopen(fname,"r"))) {printf("%s (%s)\n",strerror(errno),fname); return 1;}
while (!feof(fh))
    {
    line++;
    memset(buffer,0,1024);
    fgets(buffer,1023,fh);
    if(NULL!=(bp=strchr(buffer,'\n'))) *bp=0;
    if(NULL!=(bp=strchr(buffer,'\r'))) *bp=0;
    bp=trim(buffer);
    if(strlen(bp)==0) continue;
    if(strncmp(bp,"//",2)==0) continue;
    lineend=*(bp+strlen(bp)-1);
    defmode=0;
    if (*bp=='#') {bp=trim(++bp); defmode=1;}
    if (defmode) {
	if (strncmp(bp,"ifdef",5)==0) {bp=getdefname(bp+5,&ifdefname); if(seek_def(*def,ifdefname)) push_if(1); else push_if(0); continue;}
	if (strncmp(bp,"ifndef",6)==0) {bp=getdefname(bp+6,&ifdefname); if(seek_def(*def,ifdefname)) push_if(0); else push_if(1); continue;}
	if (strncmp(bp,"else",4)==0) {invert_if(); continue;}
	if (strncmp(bp,"endif",5)==0) {pop_if(); continue;}
	}
    if (!isif()) continue;
    if (defmode && strncmp(bp,"define",6)==0) {endmode=0; msgline=line; bp=getdefname(bp+6,&defname); mode=MODE_DEFINE;}
	else if (strncmp(bp,"static char help_text[]=",26)==0) {endmode=1; msgline=line;
	bp=bp+26; defname=strdup("MSGTR_help_text"); mode=MODE_DEFINE;}
    switch (mode) {
	case MODE_DEFINE: proc_str(bp,msgid,line,*def); break;
	}
    if (defname && endmode && lineend==';') {*def=store_def(*def,defname,msgid,msgline); free(defname); defname=NULL; memset(msgid,0,8192);}
    if (defname && !endmode && lineend!='\\') {*def=store_def(*def,defname,msgid,msgline); free(defname); defname=NULL; memset(msgid,0,8192);}
    }
if (mode==MODE_DEFINE && defname) *def=store_def(*def,defname,msgid,msgline);
if (defname) free(defname);
memset(msgid,0,8192);
fclose(fh);
return 0;
}

int main(int argc, char **argv)
{
FILE *fhout=NULL;
msgitem *item;
msgitem *fitem;
vptr *items;
int i,j, oflg=0;
long int offset;
char *finname=NULL;
char *foutname=NULL;
char *fmastername=NULL;
defitem *incdef=NULL;
defitem *tmpdef;
char *msg_id;

for(i=1;i<argc;i++) {
    if(oflg==1 && !foutname) {foutname=argv[i]; oflg=0; continue;}
    if(oflg==2 && !fmastername) {fmastername=argv[i]; oflg=0; continue;}
    if(oflg==3) {incdef=store_def(incdef,argv[i],NULL,0); oflg=0; continue;}
    if(oflg==4) {master_def=store_def(master_def,argv[i],NULL,0); trans_def=store_def(trans_def,argv[i],NULL,0); oflg=0; continue;}
    if(strcmp(argv[i],"-o")==0 && strlen(argv[i])==2) oflg=1;
    if(strcmp(argv[i],"-m")==0 && strlen(argv[i])==2) oflg=2;
    if(strcmp(argv[i],"-i")==0 && strlen(argv[i])==2) oflg=3;
    if(strcmp(argv[i],"-D")==0 && strlen(argv[i])==2) oflg=4;
    if(oflg==0 && argv[i][0]!='-' && !finname) finname=argv[i];
    }
if (!fmastername || !foutname || !finname)
    {
    printf("MPlayer message localize translator\n");
    printf("Usage: %s -o <mtm file name> -m <master file> [-i <include file> [-i <include file>]] [-D <define>] <translate file>\n",argv[0]);
    printf("\t-o:\toutput translated message file\n");
    printf("\t-m:\tmaster c header file\n");
    printf("\t-i:\tincluded c header file's\n");
    printf("\t-D:\tc defined\n");
    printf("\n");
    printf("Example: %s -o hu.mtm -m help_mp-en.h -i ../config.h -i ../libavutil/common.h -D HELP_MP_DEFINE_STATIC help_mp-hu.h\n");
    exit(1);
    }
tmpdef=incdef;
while (tmpdef) {
    proc_include(tmpdef->defname,&master_def);
    tmpdef=tmpdef->next;
    }
if (proc_include(fmastername,&master_def)) exit(1);
tmpdef=incdef;
while (tmpdef) {
    proc_include(tmpdef->defname,&trans_def);
    tmpdef=tmpdef->next;
    }
if (proc_include(finname,&trans_def)) exit(1);

tmpdef=trans_def;
while (tmpdef) {
    if (strncmp(tmpdef->defname,"MSGTR",5)==0) {
	if (NULL!=(msg_id=seek_def(master_def,tmpdef->defname))) {
	    strncpy(msgid,msg_id,8192);
	    strncpy(msgstr,tmpdef->defvalue,8192);
	    proc_msgid(tmpdef->line,tmpdef->defname);
	    } else {
	    printf("Skip definition: %s (line: %i). Don't found in master file.\n",tmpdef->defname,tmpdef->line);
	    }
	}
    tmpdef=tmpdef->next;
    }
if (itemno==0) exit(1);
item=rootitem;
items=malloc(sizeof(vptr)*itemno);
i=0;
while (item!=NULL) {
    items[i]=item;
    i++;
    item=item->next;
    }
// sort
for(i=0;i<itemno-1;i++)
    for(j=i+1;j<itemno;j++)
	{
	item=items[i];
	fitem=items[j];
	if(item->hash>fitem->hash) {items[i]=fitem; items[j]=item;}
	}
// write header
if (foutname)
    if (NULL==(fhout=fopen(foutname,"w"))) {printf("%s (%s)\n",strerror(errno),foutname);}
t_mp_trans_msg_header mp_trans_msg_header;
memset(&mp_trans_msg_header,0,sizeof(t_mp_trans_msg_header));
offset=0;
if (fhout) {
    strcpy((char*)&(mp_trans_msg_header.file_magic),FILE_MAGIC);
    strcpy((char*)&(mp_trans_msg_header.codebase),"UTF8");
    mp_trans_msg_header.file_version=0;
    mp_trans_msg_header.file_subversion=1;
    mp_trans_msg_header.index_number=itemno;
    mp_trans_msg_header.index_offset=sizeof(t_mp_trans_msg_header);
    mp_trans_msg_header.trans_start=sizeof(t_mp_trans_msg_index)*itemno+sizeof(t_mp_trans_msg_header);
    fwrite(&mp_trans_msg_header,1,sizeof(t_mp_trans_msg_header),fhout);
    }
t_mp_trans_msg_index mp_trans_msg_index;
for(i=0;i<itemno;i++) {
    memset(&mp_trans_msg_index,0,sizeof(t_mp_trans_msg_index));
    item=items[i];
    mp_trans_msg_index.text_length=strlen(item->msgid);
    mp_trans_msg_index.text_hash=item->hash;
    mp_trans_msg_index.trans_offset=offset;
    mp_trans_msg_index.trans_length=strlen(item->msgstr);
    if(fhout) {
	fwrite(&mp_trans_msg_index,1,sizeof(t_mp_trans_msg_index),fhout);
	}
    offset+=strlen(item->msgstr)+1;
    }
for(i=0;i<itemno;i++) {
    item=items[i];
    if(fhout) {
	fwrite(item->msgstr,strlen(item->msgstr)+1,1,fhout);
	}
    free(item->msgid);
    free(item->msgstr);
    free(item);
    }
printf("Process %i messages\n",itemno);
return 0;
}
